/*#######################################################################################

3c5x9.c 
  Verison 0.03   vom 16.01.2006       I.Busker  - Anpassung an weiterer 3C5x9 ID, Fehler-Rckmeldung 
  Version 0.02   vom 18.05.2005     wb  Padding und berlauf muss nicht gelesen werden.
  Version 0.01   vom 05.05.05 0:50   ;-)

Ansteuerung einer 3com 3c5x9 Netzwerkkarte fr den ...

AVR Small Webserver

Copyright (C) 2005 Werner Bllmann (wb)
Portions Copyright (C) 2004 Ulrich Radig

Dieses Programm ist freie Software. Sie knnen es unter den Bedingungen der
GNU General Public License, wie von der Free Software Foundation verffentlicht,
weitergeben und/oder modifizieren, entweder gem Version 2 der Lizenz oder
(nach Ihrer Option) jeder spteren Version.

Die Verffentlichung dieses Programms erfolgt in der Hoffnung,
da es Ihnen von Nutzen sein wird, aber OHNE IRGENDEINE GARANTIE,
sogar ohne die implizite Garantie der MARKTREIFE oder der VERWENDBARKEIT
FR EINEN BESTIMMTEN ZWECK. Details finden Sie in der GNU General Public License.

Sie sollten eine Kopie der GNU General Public License zusammen mit diesem
Programm erhalten haben.
Falls nicht, schreiben Sie an die Free Software Foundation,
Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

#######################################################################################*/
/* ************************************************************************************
Auf dem AtMega32 Board von Holger Buss. 

Auftrennen der Verbindung von A9 mit A8 (ISA Slot Pin A22-A23, Ltseite)
und verbinden mit A4 (ISA Slot Pin A27, Ltseite).
Wenn A9 mit A4 verbunden ist, dann ist der
ID-Port der 3c5x9 = 0x100 und der Datenport = 0x310

Ansprechen des IDPorts ber Adresse 0x00.
Ansprechen der Datenregister ber Adresse 0x10-0x1F.

Die Karte muss vorher mit 3c5x9cfg.exe konfiguriert sein.
Aktuelle Einstellungen: INT3, I/O 310H, TP, PnP aus, Full-Duplex aus.
*/

#include <stdio.h>
#include <string.h>

#include "main.h"

#if USE_3C5x9

#include <avr/io.h>


// ********************* MACROS ***************************
#if defined (__AVR_ATmega32__)
//Ethernet Card Ports

/*#######################################################################################
// Belegung fr Holger Buss Mega32 WebBoard
#######################################################################################*/


/*
 * Steuerleitungen
 */
#define CTRL_DDR        DDRB
#define CTRL_OUT        PORTB
#define CTRL_RD         _BV(PB2)
#define CTRL_WR         _BV(PB1)
#define CTRL_RESET      _BV(PB0)

#define RESET_ACTIVE()  (CTRL_OUT |=  CTRL_RESET)
#define RESET_PASSIVE() (CTRL_OUT &= ~CTRL_RESET)
#define RD_ACTIVE()     (CTRL_OUT &= ~CTRL_RD)
#define RD_PASSIVE()    (CTRL_OUT |=  CTRL_RD)
#define WR_ACTIVE()     (CTRL_OUT &= ~CTRL_WR)
#define WR_PASSIVE()    (CTRL_OUT |=  CTRL_WR)

/*
 * Adressleitungen A0-A4
 */
#define ADDRESS_DDR     DDRA
#define ADDRESS_OUT     PORTA
#define ADDRESS_MASK    0x1F
#define PORT_REGISTER(a) ADDRESS_OUT=(ADDRESS_OUT&~ADDRESS_MASK)|(a & ADDRESS_MASK)

/*
 * Datenleitungen
 */
#define DATA_DDR        DDRC
#define DATA_OUT        PORTC
#define DATA_IN         PINC
#define DATA_TO_OUT()   (DATA_DDR = 0xFF)
#define DATA_TO_IN()    (DATA_DDR = 0x00)
#else
  #error PORT for your MCU not defined yet.
#endif


// Die wichtigsten  IO-Port Adressen als Makro
#define ID_PORT  0x100
#define EEDATA_L 0x31C
#define EEDATA_H 0x31D
#define STATUS   0x31E

#define delay_100ms() delay(100)

// 3c5x9 control macros
#define select_register_window(_window)	    if(1){_3c5x9_window__=_window&7;setreg(0x31E,0x0800|_3c5x9_window__);}
#define global_reset()	                    setreg(0x31E, 0x0000)
#define start_coax_tranceiver()             setreg(0x31E, 0x1000)
#define rx_disable()                        setreg(0x31E, 0x1800)
#define rx_enable()                         setreg(0x31E, 0x2000)
#define rx_reset() 	                        setreg(0x31E, 0x2800)
#define rx_discard_top_packet()   	        setreg(0x31E, 0x4000)
#define tx_enable()   	                    setreg(0x31E, 0x4800)
#define tx_disable()  	                    setreg(0x31E, 0x5000)
#define request_interrupt()   	            setreg(0x31E, 0x6000)
#define statistics_enable()   	            setreg(0x31E, 0xA800)
#define statistics_disable()                setreg(0x31E, 0xB000)
#define stop_coax_transceiver()  	        setreg(0x31E, 0xB800)
#define power_down_full()  	                setreg(0x31E, 0xE000)
#define power_auto()  	                    setreg(0x31E, 0xE800)

#define set_interrupt_mask(mask) 	        setreg(0x31E, 0x7000 | (mask & 0xFF))
#define set_read_zero_mask(mask)	        setreg(0x31E, 0x7800 | (mask & 0xFF))
#define set_rx_early_threshold(ths)	        setreg(0x31E, 0x8800 | (ths & 0x7FF))
#define set_tx_available_threshold(ths) 	setreg(0x31E, 0x9000 | (ths & 0x7FF))
#define set_tx_start_threshold(ths)	        setreg(0x31E, 0x9800 | (ths & 0x7FF))
#define set_tx_claim_threshold(ths)  	    setreg(0x31E, 0xC000 | (ths & 0x7FF))

// After an underrun or jabber error on transmit, a TX Reset is required.
// Do not use TX Reset unless it is absolutely necessary.
#define  tx_reset()  	                    setreg(0x31E,0x5800)

// Interrupt reasons:
// 0x01 = Interrupt Latch
// 0x02 = Adapter Failure
// 0x04 = TX Complete
// 0x08 = TX Available
// 0x10 = RX Complete
// 0x20 = RX Early
// 0x40 = Interrupt Request
// 0x80 = Update  Statistics
#define ack_interrupt(reason)	            setreg(0x31E,0x6800|(reason&0xFF))

// RX filter mask values
// 0x01 = Individual address
// 0x02 = Multicast address
// 0x04 = broadcast address
// 0x08 = promiscuous mode
#define set_rx_filter(filter)	            setreg(0x31E,(0x8000|(filter&0x0F)))



// ***************** DATA *************************
/*
 * Hilfsvariable fr delay()
 */
static uint16_t delay_count = F_CPU / 4000;

/*
 * Hier wird das aktuell ausgewaehlte Register-Fenster gemerkt fr den
 * Fall dass eine Interruptroutine sie ndert und wieder zurckstellen will.
 */
uint8_t _3c5x9_window__ = 0;

/*
 * Sendedaten Kopfinformation
 */
static uint8_t tx_preamble[] =
{
	14+80, 0x00, 0x00, 0x00     // TRUE_PACKET_LENGTH, 0, 0, 0
};

// ******************* CODE ********************
/*
 * Ein byte ausgeben
 */
void outportb(uint16_t p, uint8_t b)
{
    PORT_REGISTER(p);
    DATA_OUT    = b;
    DATA_TO_OUT();
    WR_ACTIVE();        // kein nop beim schreiben ntig
nop();
nop();
    WR_PASSIVE();
    DATA_TO_IN();
}

/*
 * Ein byte einlesen
 */
uint8_t inportb(uint16_t p)
{
    PORT_REGISTER(p);
    RD_ACTIVE();
    nop();          // Minimum 2 nop's beim lesen
    nop();
    p = DATA_IN;
    RD_PASSIVE();
    return p;
}


/*
 * Ein 2 byte wort ausgeben
 */
// "...commands must be written as a word, ... (executed when the high byte is written)..."

// Die reihenfolge des schreibens der register ist in "Chapter 6" ganz am anfang des datenblatts
// definiert. "...the low byte acceess must be followed by the high byte access,..."
void setreg(uint16_t reg, uint16_t val)
{
    DATA_OUT    = val;          // low byte
    reg &= ADDRESS_MASK;
    reg |= (ADDRESS_OUT & ~ADDRESS_MASK);
    ADDRESS_OUT = reg;
    DATA_TO_OUT();
    reg++;
    WR_ACTIVE();
    val >>= 8;
    WR_PASSIVE();
    nop();
    nop();
    DATA_OUT    = val;          // high byte
    ADDRESS_OUT = reg;
    WR_ACTIVE();
nop();
nop();
    WR_PASSIVE();
    DATA_TO_IN();
}

/*
 * Ein 2 byte wort einlesen
 */
// Die reihenfolge des lesens der register ist in "Chapter 6" ganz am anfang des datenblatts
// definiert. "...the low byte acceess must be followed by the high byte access,..."
uint16_t readreg(uint16_t reg)
{
    uint16_t st;

    reg &= ADDRESS_MASK;
    reg |= (ADDRESS_OUT & ~ADDRESS_MASK);
    ADDRESS_OUT = reg;
    RD_ACTIVE();
    reg++;
    nop();
nop();
    st = DATA_IN;   // lower byte
    RD_PASSIVE();
    nop();
    nop();
nop();
    ADDRESS_OUT = reg;
    RD_ACTIVE();
    nop();
    nop();
nop();
    st |= (DATA_IN << 8); // upper byte
    RD_PASSIVE();
    return st;
}


/*
 * Hilfsroutine (Aus WinAVR LibC documentation)
 */
void delay(uint8_t ms)
{
    uint16_t cnt;
    asm volatile (
        "\n"
        "L_dl1%=:" "\n\t"
        "mov %A0, %A2" "\n\t"
        "mov %B0, %B2" "\n"
        "L_dl2%=:" "\n\t"
        "sbiw %A0, 1" "\n\t"
        "brne L_dl2%=" "\n\t"
        "dec %1" "\n\t"
        "brne L_dl1%=" "\n\t"
        : "=&w" (cnt)
        : "r" (ms), "r" (delay_count)
        );
}


#if defined (DEBUG)
/*
 *  Degug hilfe um prozessor temporr zu stoppen
 */
int16_t getch(void)
{
    int16_t ch;

    UCR |= (1 << RXEN);  // RX enable
    while(!(UCSRA & _BV(RXC)))
        ;
    ch = UDR;
    UCR &= ~(1 << RXEN);  // RX disable
    return ch;
}
#endif

/*
 * kein macro
 */
void power_up(void)
{
  	 setreg(0x31E, 0xD800);
     tx_reset(); // see Page 39 of 3c5x9 manual
}

/*
 * Lese Adaper-Konfinguartionsinformationen aus dem Adapter eeProm
 */
uint16_t readAdapterEEProm(uint16_t reg)
{
 	reg &= 0x3F;	// nur die unteren 6 bits gltig
    select_register_window(0);
	// Warte wenn eeProm busy
	while(inportb(0x31B) & 0x80)
		;		// EEProm busy
	setreg(0x31A, 0x80 + reg);
	// Warte wenn eeProm busy
	while(inportb(0x31B) & 0x80)
		;		// EEProm busy
	return readreg(0x31C);
}


/*
 * Die netzwerkkarte muss erst "aktiviert" werden damit sie, und nur sie auf dem
 * port 0x1x0 befehle entgegennimmt und informationen zurckgibt.
 * dazu muss ein bestimmtes bytemuster auf den port (ID_PORT) im bereich
 * 0x1x0 (x=1...F) ausgegeben werden, auf den als letztes der wert null
 * geschrieben wurde.
 */
static void activate_3c5x9_nic(void)
{
	uint16_t id;
    uint16_t loop;
    // ----------------
    // Write ID loop
    id   = 0xFF;
	outportb(ID_PORT, 0);			// Enable IDport
	outportb(ID_PORT, 0);			// Enable IDport
	for(loop=0; loop < 255; loop++)
	{
		outportb(ID_PORT, id);
		id <<= 1;
		if(id & 0x100)
			id ^= 0xCF;
	}
}

/*
 * Entspricht einem hardware-reset
 */
void nic_global_reset(void)
{
    // Write ID loop
    activate_3c5x9_nic();
    // do a global reset
    outportb(ID_PORT, 0xCF);
    delay(127);             // Warte bis reset ausgefrhrt
}

/*
 * Um von ID_PORT eeprom-daten zu lesen mssen diese 16-bit werte nach anforderung in
 * einer schleife vom bit #0 des ID_PORT eingelsen werden. Hchstwertiges bit zuerst.
 */
uint16_t readLoop(uint8_t eadr)
{
    uint8_t loop;
    uint16_t st = 0;  // to keep the compiler happy

    // Fordere controller auf die daten von eeprom adresse eadr zu lesen.
    outportb(ID_PORT, 0x80+eadr);
    delay(100);     // Warte bis daten im eeprom datenregister zur verfgung stehen

	for(loop = 0; loop < 16; loop++)
    {
   	    st <<= 1;
		st |= inportb(ID_PORT) & 1;
        nop();      // hat sich bewhrt ;-)
    }
    return st;
}

/*
 * Initialisieren der datenports des ATmega und der 3c5x9
 * Liefert 0 wenn erfolgreich, !=0 sonst.
 */
int init_3c5x9(void)
{
	uint16_t st;

printf("Init...\n");

printf_P(PSTR("Phase 1\n"));

 
    // Steuerleitungen
    // RD + WR Passive, Reset Active
    CTRL_OUT        |= (CTRL_RD | CTRL_WR | CTRL_RESET);
    CTRL_DDR        |= (CTRL_RD | CTRL_WR | CTRL_RESET);
    // reset ist momentan aktiv

    // Address lines
    ADDRESS_DDR     |= ADDRESS_MASK;

    // Addressiert wird 0x100
    PORT_REGISTER(0x100);
    DATA_TO_IN();

printf_P(PSTR("Phase 2\n"));
    delay(100);         // Reset halten
    RESET_PASSIVE();
printf_P(PSTR("Phase 3\n"));

    // Erst mal warten, nic initialisiert sich
    delay(100);

printf_P(PSTR("Phase 4\n"));

    // Aktivierungssequenz schicken
	activate_3c5x9_nic();

printf_P(PSTR("Phase 5\n"));

	// read the mac address of this card from eeprom address 10,11, and 12
    st = readLoop(10);
    printf_P(PSTR("node %02X-"),st);

    // Every nic i have seen has "00" in the msb mac field
    if((st & 0xff00) != 0)      // no nic, i guess
   	    return 1;

	mymac[0] = st>>8;
	mymac[1] = st;
    st = readLoop(11);
    printf_P(PSTR("%02X-"),st);
	mymac[2] = st>>8;
	mymac[3] = st;
    st = readLoop(12);
    printf_P(PSTR("%02X\n"),st);
	mymac[4] = st>>8;
	mymac[5] = st;

    // --------------------
    // read the I/O configuration from eepro address 8
    st = readLoop(8);
    printf_P(PSTR("I/O address %02X\n"),(st<<4)+0x200);
#if defined (NAGGING_CHECK)
    // Prfung nicht unbeding notwendig, da die I/O spter noch
    // per software (ID_PORT) gendert werden kann
    if(st != ((0x310-0x200)>>4))  // 0x11
        return 2;
#endif
//getch();
printf_P(PSTR("Phase 6\n"));

    // Setze basis I/O Adresse auf 0x310
	// Activate the adapter at the preconfigured I/O base address and return to
    // the ID_WAIT state.
    outportb(ID_PORT, 0xE0+((0x310-0x200)>>4));  // 0F1H fr 310H
    outportb(ID_PORT, 0x0);
printf_P(PSTR("Phase 7\n"));

    // select_register_window(0);  // nach setzten von I/O addresse sowiso auf window 0
    // Warte solange karte "busy"
    while((st = readreg(0x31E)) & 0x1000)
        printf("a %04X\r", st);

    // Hole einige informationen von der karte
    st = readreg(0x310);       // Hersteller Kennung
    printf("MfgtID: %04X\n", st);
    if(st != 0x6D50)           // 3Com ?
    { printf_P(PSTR("unknown ID\n")); 	return 3; }   // no 3com

    // Produktkennung
	st = readreg(0x312);
    printf_P(PSTR("Product ID: %04X\n"), st);
    if((st != 0x9450) && (st!=0x9550))
    	{ printf_P(PSTR("unknown ID\n")); //return 4; 
		}   // no 3c5x9

#if defined (DEBUG)
    st = readreg(0x314);
    printf_P(PSTR("Configuration Control: %04X\n"), st);

    st = readreg(0x316);
    printf("Address configuration: %04X\n", st);

    st = readreg(0x318);
    printf_P(PSTR("Resource configuration: %04X\n"), st);


    st = readreg(STATUS);
    printf_P(PSTR("Status + Window: %04X\n"), st);

    // eeprom komlett auslesen
    for(i = 0; i < 0x40; i++)
    {
   	    st = readAdapterEEProm(i);
   	    if((i % 6) == 0)
   		    printf_P(PSTR("\nEEProm: "));
   	    printf_P(PSTR("%02d = %04X, "), i, st);
    }
printf_P(PSTR("\nPhase 8\n"));
    getch();
#endif // DEBUG
printf_P(PSTR("Phase 9\n"));
    /*
     * Set the mac address
     */
	// Switch to command window 2 (Station address management)
    select_register_window(2);
	outportb(0x315, mymac[5]);
	outportb(0x314, mymac[4]);
	outportb(0x313, mymac[3]);
	outportb(0x312, mymac[2]);
	outportb(0x311, mymac[1]);
	outportb(0x310, mymac[0]);
printf_P(PSTR("Phase 10\n"));

    // Enable Link Beat and Jabber
    select_register_window(4);
    st = readreg(0x31A);
    st |= (0x80 /* link beat */ | 0x40 /* jabber */);
    setreg(0x31A, st);
    set_read_zero_mask(0xFF); //P40 (6-8)
    set_interrupt_mask(0xFF); 

    //power_auto();

	// Sender enable
    tx_reset();   // danach ne kleine pause
    delay(100);
    do
    	st = readreg(0x31E);
    while(st & 0x1000);

    /*
     * Empfangs paketfilter setzten
     */
    //set_rx_filter(1);         // hre nur auf eigene addresse
    //set_rx_filter(0x08);      // promiscuous = Alles!

    // hre nur auf eigene adresse und auf broadcasts (0xFFFFFFFFFFFF)
    set_rx_filter(0x05);   // broadcast + individual
printf_P(PSTR("Phase 11\n..."));

    // Sender und Empfnger aktivieren
    tx_enable();
    rx_enable();

	return 0;           // That's it!
}

// ################################################
//############################################################################
//void Write_Ethernet_Frame (struct Ethernet_Header *buffer,int16_t bufferlen)
void Write_Ethernet_Frame (uint8_t *buffer,int16_t bufferlen)
//############################################################################
{
	uint16_t     j;
	uint8_t      v;

	select_register_window(1);

    v = inportb(0x31B);	// TX Status
    if(v & 0x60)
    {
        printf_P(PSTR("TX Status: Jabber or Underrung - TX reset \n"));
        tx_reset();
        delay(1);
        // und warte bis mit reset fertig
        do
        	v = readreg(0x31E);
        while(v & 0x1000);

        tx_enable();
    }

    // Wait for buffer space
    *((unsigned *)&tx_preamble[0]) = bufferlen;     // Paketlnge
    do
    {
   	    j = readreg(0x31C);   // free tx buffer bytes
    }
    while(j < (bufferlen+14));

    // Write the preamble
    PORT_REGISTER(0x310);   // DMA port in window 1
    DATA_TO_OUT();          // we want to write to the adapter

    j = 0;
    while(j < sizeof(tx_preamble))
    {
		DATA_OUT = tx_preamble[j];
        WR_ACTIVE();
        j++;
        WR_PASSIVE();
    }

    // Write the packet data
    j = 0;
    while(j < bufferlen)
    {
		DATA_OUT = ((char *)buffer)[j];
        WR_ACTIVE();
        j++;
        WR_PASSIVE();
    }

    // Write padding
	DATA_OUT = 0x55;
    while(j & 3)
    {
        WR_ACTIVE();
        j++;
        WR_PASSIVE();
    }
    DATA_TO_IN();
	DATA_OUT = 0;
}

//############################################################################
//void Read_Ethernet_Frame (struct Ethernet_Header *buffer,int16_t *bufferlen)
void Read_Ethernet_Frame (uint8_t *buffer,int16_t *bufferlen)
//############################################################################
{
	uint16_t i, j, k, v;

 	select_register_window(1);

    /* Lese Packet */
    i = 0;
    j = 0;
    v = readreg(0x318);             // RX status
    k = v & 0x7FF;                  // isoliere Lnge
    do {
        if(v & 0x4000)              // error in RX packet
        {
            rx_discard_top_packet();// wechschmeien

            char ca[40] = "";
            v &= 0x3800;	        // isoliere Fehlerinformation
            if(v == 0)
                strcat_P(ca, PSTR("Overrun "));
            else if(v == 0x1800)
                strcat_P(ca, PSTR("Runt packet error "));
            else if(v == 0x2000)
                strcat_P(ca, PSTR("Alignment error "));
            else if(v & 0x2800)
                strcat_P(ca, PSTR("CRC error "));
            else if(v & 0x800)
                strcat_P(ca, PSTR("Oversized "));
            printf_P(PSTR("\n =====> %s<====\n"), ca);
            *bufferlen = 0;
            return;
        }
        // Lnge aufaddieren falls mglich
        if(k && (k != 0x7FF) && (k != 0x7FE))
            j += k;

        // DMA port in window 1
        PORT_REGISTER(0x310);   // war durch readreg gendert!
        for(; (i < j) && (i < MTU_SIZE); i++)
        {
            RD_ACTIVE();
            nop();
            nop();
            ((char *)buffer)[i] = DATA_IN;
            RD_PASSIVE();
        }
        v = readreg(0x318);     // RX status
        k = v & 0x7FF;          // isoliere Lnge
        // solange packet nicht vollstndig eingetroffen/gelesen
    } while (((v & 0x8000) || (k && (k != 0x7FF) && (k != 0x7FE))) && (i < MTU_SIZE));

    // nur die wirklich gelesene lnge zurckliefern
    *bufferlen = i;
#if 0
    // Daten abholen die NICHT mehr in den Buffer passen
    //  && padding bytes
    PORT_REGISTER(0x310);   // war durch readreg gendert!
    while((i < j) || (i & 3))
	{
        RD_ACTIVE();
        nop();
        nop();
        v = DATA_IN;
        RD_PASSIVE();
        i++;
	}
#endif	
    rx_discard_top_packet();  // Packet erledigt !
}

/*
 * Adapter polling routine
 * Wird in der main-loop immer wieder augerufen
 */
void do_network_io(void)
{
#if 1
    select_register_window(1);          // 
    if((inportb(0x319) & 0x80) == 0)    // Wenn "RX Status incomplete" gelscht
    {
    	get ();
    }
#else
    uint16_t st;

    select_register_window(4);  // Status-Fenster
    if((st = readreg(STATUS)) & 0x10)  // Receive complete
    {
        printf("+packet %04X\n", st);
    	get ();
        printf("-packet\n");
    }
#endif    
}
#endif
